from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_two_alternatives():
    regex = r'a|b'
    possibilities = ['a', 'b']

    f_finite(regex, possibilities)


def test_alternatives_with_quantifier_on_second_option():
    regex = r'a|b*'
    possibilities = ['a', '', 'b', 'bb', 'bbb', 'bbbb', 'bbbbb']

    f_infinite(regex, possibilities)


def test_alternatives_with_quantifier_plus_on_first_option():
    regex = r'a+|b'
    possibilities = ['b', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regex, possibilities)


def test_multiple_alternatives():
    regex = r'a|b|c'
    possibilities = ['a', 'b', 'c']

    f_finite(regex, possibilities)


def test_alternative_with_literal_and_character_class():
    regex = r'a|[b-d]'
    possibilities = ['a', 'b', 'c', 'd']

    f_finite(regex, possibilities)


def test_alternative_with_character_class_and_literal():
    regex = r'[a-c]{ 0}|d'
    possibilities = ['', 'd']

    f_finite(regex, possibilities)


def test_alternation_with_character_classes_and_literals():
    regex = r'(a|[0-2])'
    possibilities = ['a', '0', '1', '2']

    f_finite(regex, possibilities)


def test_nested_alternation():
    regex = r'((a|b)|c)'
    possibilities = ['a', 'b', 'c']

    f_finite(regex, possibilities)


def test_alternation_with_grouping():
    regex = r'(a(b|c)d|x)'
    possibilities = ['abd', 'acd', 'x']

    f_finite(regex, possibilities)


def test_same_alternative_twice():
    regex = r'a{1,2}|a{1,2}'
    possibilities = ['a', 'aa']

    f_finite(regex, possibilities)



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_empty_pattern_yields_empty_string():
    regex = r''
    possibilities = ['']
    f_finite(regex, possibilities)


def test_single_literal_character():
    regex = r'a'
    possibilities = ['a']
    f_finite(regex, possibilities)


def test_zero_or_more_quantifier_on_single_char():
    regex = r'a*'
    possibilities = ['', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']
    f_infinite(regex, possibilities)


def test_one_or_more_quantifier_on_single_char():
    regex = r'a+'
    possibilities = ['a', 'aa', 'aaa', 'aaaa', 'aaaaa']
    f_infinite(regex, possibilities)


def test_zero_or_one_quantifier_on_single_char():
    regex = r'a?'
    possibilities = ['', 'a']
    f_finite(regex, possibilities)


def test_exact_repetition_quantifier_on_single_char():
    regex = r'a{2}'
    possibilities = ['aa']
    f_finite(regex, possibilities)


def test_minimum_repetition_quantifier_on_single_char():
    regex = r'a{2,}'
    possibilities = ['aa', 'aaa', 'aaaa', 'aaaaa']
    f_infinite(regex, possibilities)


def test_min_max_repetition_quantifier_on_single_char():
    # `a{2,4}` yields 'aa', 'aaa', 'aaaa'.
    regex = r'a{2,4}'
    possibilities = ['aa', 'aaa', 'aaaa']
    f_finite(regex, possibilities)


def test_zero_times_repetition_quantifier_on_single_char():
    regex = r'a{0}'
    possibilities = ['']
    f_finite(regex, possibilities)


def test_escaped_literal_special_characters():
    regex = r'\*\+\?'
    possibilities = ['*+?']
    f_finite(regex, possibilities)


def test_single_character_class():
    regex = r'[abc]'
    possibilities = ['a', 'b', 'c']
    f_finite(regex, possibilities)


def test_single_escaped_character():
    regex = r'\n'
    possibilities = ['\n']
    f_finite(regex, possibilities)


def test_literal_dot_character():
    regex = r'\.'
    possibilities = ['.']
    f_finite(regex, possibilities)



from regex_enumerator import RegexEnumerator


def f_test(regexEnumerator: RegexEnumerator, possibilities: list[str]):
    while len(possibilities) != 0:
        res = regexEnumerator.next()
        assert res in possibilities, f"'{res}' is not in {possibilities}"
        possibilities.remove(res)


def f_finite(regex: str, possibilities: list[str], additional_charset: str = ''):
    regexEnumerator = RegexEnumerator(
        regex, additional_charset=additional_charset)
    possibilities1 = possibilities.copy()
    f_test(regexEnumerator, possibilities)
    assert regexEnumerator.next() is None
    assert regexEnumerator.done

    regexEnumerator = RegexEnumerator(
        regex, additional_charset=additional_charset, precompute=False)
    f_test(regexEnumerator, possibilities1)
    assert regexEnumerator.next() is None
    assert regexEnumerator.done


def f_infinite(regex: str, possibilities: list[str], additional_charset: str = ''):
    regexEnumerator = RegexEnumerator(
        regex, additional_charset=additional_charset)
    possibilities1 = possibilities.copy()
    f_test(regexEnumerator, possibilities)
    assert not regexEnumerator.done

    regexEnumerator = RegexEnumerator(
        regex, additional_charset=additional_charset, precompute=False)
    f_test(regexEnumerator, possibilities1)
    assert not regexEnumerator.done



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_not_capturing_groups():
    regex = r'(?:a)(b)\1'
    possibilities = ['abb']

    f_finite(regex, possibilities)



from regex_enumerator import RegexEnumerator
from .test_function import f_finite


def test_digit_escape():
    regex = r'\d'
    possibilities = [str(i) for i in range(10)]

    f_finite(regex, possibilities)


def test_digit_escape_with_quantifier():
    regex = r'\d{1 , 2 }'
    possibilities = [str(i) for i in range(10)] + [str(i) + str(j)
                                                   for i in range(10) for j in range(10)]

    f_finite(regex, possibilities)


def test_non_digit_escape():
    regex = r'\D'
    possibilities = [chr(i)
                     for i in range(32, 127) if chr(i) not in '0123456789']

    f_finite(regex, possibilities)


def test_word_escape():
    regex = r'\w'
    possibilities = [chr(i) for i in range(
        32, 127) if chr(i).isalnum() or chr(i) == '_']

    f_finite(regex, possibilities)


def test_non_word_escape():
    regex = r'\W'
    possibilities = [chr(i) for i in range(
        32, 127) if not (chr(i).isalnum() or chr(i) == '_')]

    f_finite(regex, possibilities)


def test_whitespace_escape():
    regex = r'\s'
    possibilities = [' ', '\t', '\n', '\r', '\f', '\v']

    f_finite(regex, possibilities)


def test_non_whitespace_escape():
    regex = r'\S'
    possibilities = [chr(i) for i in range(
        32, 127) if chr(i) not in ' \t\n\r\f\v']

    f_finite(regex, possibilities)


def test_tab_escape():
    regex = r'\t'
    possibilities = ['\t']

    f_finite(regex, possibilities)


def test_carriage_return_escape():
    regex = r'\r'
    possibilities = ['\r']

    f_finite(regex, possibilities)


def test_newline_escape():
    regex = r'\n'
    possibilities = ['\n']

    f_finite(regex, possibilities)


def test_vertical_tab_escape():
    regex = r'\v'
    possibilities = ['\v']

    f_finite(regex, possibilities)


def test_form_feed_escape():
    regex = r'\f'
    possibilities = ['\f']

    f_finite(regex, possibilities)


def test_hex_escape():
    regex = r'\x41'
    possibilities = ['A']

    f_finite(regex, possibilities)


def test_escaped_open_square_bracket():
    regex = r'\['
    possibilities = ['[']

    f_finite(regex, possibilities)


def test_escaped_open_close_square_brackets():
    regex = r'\[\]'
    possibilities = ['[]']

    f_finite(regex, possibilities)


def test_escaped_characters_inside_character_class():
    regex = r'[\[\]]'
    possibilities = ['[', ']']

    f_finite(regex, possibilities)


def test_escaped_char_interrups_range_after_divider():
    regex = r'[a-\d]'
    possibilities = ['a', '-', '0', '1', '2',
                     '3', '4', '5', '6', '7', '8', '9']

    f_finite(regex, possibilities)


def test_escaped_char_interrups_range_after_1st_char():

    regex = r'[\[\d]'
    possibilities = ['[', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9']

    f_finite(regex, possibilities)


def test_escaped_unicode_literal():
    regex = r'\u00E0'
    possibilities = ['à']

    f_finite(regex, possibilities)



from regex_enumerator import RegexEnumerator
from .test_function import f_finite, f_infinite


def test_single_capturing_group_with_literal():
    regex = r'(a)'
    possibilities = ['a']

    f_finite(regex, possibilities)


def test_single_capturing_group_with_class_single_char():
    regex = r'([a])'
    possibilities = ['a']

    f_finite(regex, possibilities)


def test_single_capturing_group_with_class_multi_char():
    regex = r'([a-c])'
    possibilities = ['a', 'b', 'c']

    f_finite(regex, possibilities)


def test_capturing_group_with_star_quantifier():
    regex = r'(a)*'
    possibilities = ['', 'a', 'aa', 'aaa', 'aaaa', 'aaaaa']

    f_infinite(regex, possibilities)


def test_named_capturing_group_with_optional_subgroup():
    regex = r'(?<name>a[bcd](e)?)'
    possibilities = ['ab', 'abe', 'ac', 'ace', 'ad', 'ade']

    f_finite(regex, possibilities)


def test_literal_followed_by_group_with_star_quantifier():
    regex = r'a(b)*'
    possibilities = ['a' + 'b' * i for i in range(6)]

    f_infinite(regex, possibilities)


def test_two_capturing_groups_with_star_quantifiers():
    regex = r'(a)*(b)*'
    possibilities = ['a' * i + 'b' * j for i in range(6) for j in range(6)]

    f_infinite(regex, possibilities)


def test_nested_capturing_groups():
    regex = r'(a(b(c)))'
    possibilities = ['abc']

    f_finite(regex, possibilities)


def test_capturing_groups_in_sequence():
    regex = r'((a)(b))'
    possibilities = ['ab']

    f_finite(regex, possibilities)


def test_non_capturing_group():
    regex = r'(?:a|b)*'
    possibilities = ['', 'a', 'b', 'aa', 'ab', 'ba', 'bb']

    f_infinite(regex, possibilities)


def test_non_capturing_group_with_quantifier():
    regex = r'(?:ab)+'
    possibilities = ['ab', 'abab', 'ababab']

    f_infinite(regex, possibilities)


def test_named_capturing_group_with_quantifier():
    regex = r'(?<chars>[ab]{1,2})'
    possibilities = ['a', 'b', 'aa', 'ab', 'ba', 'bb']

    f_finite(regex, possibilities)


def test_nested_non_capturing_groups():
    regex = r'(?:a(?:b(?:c)))?'
    possibilities = ['', 'abc']

    f_finite(regex, possibilities)


def test_group_for_quantifier_scope():
    regex = r'(ab)+'
    possibilities = ['ab', 'abab', 'ababab']

    f_infinite(regex, possibilities)


def test_group_with_char_class_infinite_repetition():
    regex = r'([ab])+'
    possibilities = ['a', 'b', 'aa', 'ab', 'ba', 'bb']

    f_infinite(regex, possibilities)


def test_group_with_multiple_elements_with_qunatifiers():
    regex = r'(a[b-d]{0,2}){0, 3}'
    possibilities = ['']
    char_class = ['', 'b', 'c', 'd', 'bb', 'bc',
                  'bd', 'cb', 'cc', 'cd', 'db', 'dc', 'dd']
    one = [f'a{c}' for c in char_class]
    two = [f'{c1}{c2}' for c1 in one for c2 in one]
    three = [f'{c1}{c2}{c3}' for c1 in one for c2 in one for c3 in one]
    possibilities.extend(one)
    possibilities.extend(two)
    possibilities.extend(three)

    f_finite(regex, set(possibilities))


def test_nested_groups_with_multiple_elements_with_quantifiers():
    regex = r'(a([e-g]){1, 3}){0, 3}'
    possibilities = ['']
    group = ['e', 'f', 'g', 'ee', 'ef', 'eg', 'fe', 'ff', 'fg', 'ge', 'gf', 'gg', 'eee', 'eef', 'eeg', 'efe', 'eff', 'efg', 'ege', 'egf', 'egg',
             'fee', 'fef', 'feg', 'ffe', 'fff', 'ffg', 'fge', 'fgf', 'fgg', 'gee', 'gef', 'geg', 'gfe', 'gff', 'gfg', 'gge', 'ggf', 'ggg']
    one = [f'a{g}' for g in group]
    two = [f'{g1}{g2}' for g1 in one for g2 in one]
    three = [f'{g1}{g2}{g3}' for g1 in one for g2 in one for g3 in one]
    possibilities.extend(one)
    possibilities.extend(two)
    possibilities.extend(three)

    f_finite(regex, set(possibilities))
